#!/usr/bin/perl
################################################################################
# NParse v0.1.5 by h1kari, for Perl with Net::RawIP mods.                      #
# [http://www.nfsg.org] (Nightfall Security Group)                             #
# Usage: nparse.pl <-o<outfile>>[-c<config>]                                   #
# Simple configurable parse script for ndump.pl output files. Prints out       #
# configured protocol headers in hex, ord, and bin and payload in ascii/hex.   #
# Currently it supports arp, ip, tcp, udp, icmp, rawip. It shouldn't be too    #
# hard to configure more protocols with the way I have it setup. If you have   #
# any suggestions or comments, please send them to me.                         #
# [h1kari@nfsg.org]                                                            #
#                                                                              #
################################################################################
# Config Syntax: <prot=<protocol>> [inc='<data>'] [off=<offset>] [size=<size>] #
################################################################################
#                                                                              #
# e.g., to print only tcp packets containing the data 'pass' and 'user', and   #
#       ones that are destined to port 23 (0x17) you would do this:            #
#  $ echo "prot=tcp inc='pass|user' off=40                                     #
#  > prot=tcp inc='\x00\x17' off=22 size=2" > config                           #
#  $ chmod 755 nparse.pl                                                       #
#  $ ./nparse.pl -o output.file -c config                                      #
#                                                                              #
# NOTE: the offset always starts at the top of the hardware link offset.       #
#                                                                              #
################################################################################
# Optional Configuration Settings               # Descriptions                 #
################################################################################
my($ver)='0.1.5';                               # for version control purposes #
my($defsize)=1024;                              # default search size for pkts #
my(@prs);                                       # create array for protocols   #
################################################################################
# Define the protocol hash                                                     #
################################################################################
# First Level Protocols Here ###################################################
## IP Protocol #################################################################
$prs[0]{'id'}='ip';
$prs[0]{'di'}=0x0800;
$prs[0]{'name'}='Internet Protocol';
$prs[0]{'size'}='20';
$prs[0]{'off'}='-2';
$prs[0]{'offs'}='2';
$prs[0]{'pload'}='0';
## ARP Protocol ################################################################
$prs[1]{'id'}='arp';
$prs[1]{'di'}=0x0806;
$prs[1]{'name'}='Address Resolution Protocol';
$prs[1]{'size'}='28';
$prs[1]{'off'}='-2';
$prs[1]{'offs'}='2';
$prs[1]{'pload'}='0';
################################################################################
## Second Level Protocols Here #################################################
### TCP Protocol ###############################################################
$prs[0]{'prs'}[0]{'id'}='tcp';
$prs[0]{'prs'}[0]{'di'}='6';
$prs[0]{'prs'}[0]{'name'}='Transmission Control Protocol';
$prs[0]{'prs'}[0]{'size'}='20';
$prs[0]{'prs'}[0]{'off'}='9';
$prs[0]{'prs'}[0]{'offs'}='1';
$prs[0]{'prs'}[0]{'pload'}='1';
### UDP Protocol ###############################################################
$prs[0]{'prs'}[1]{'id'}='udp';
$prs[0]{'prs'}[1]{'di'}='17';
$prs[0]{'prs'}[1]{'name'}='User Datagram Protocol';
$prs[0]{'prs'}[1]{'size'}='8';
$prs[0]{'prs'}[1]{'off'}='9';
$prs[0]{'prs'}[1]{'offs'}='1';
$prs[0]{'prs'}[1]{'pload'}='1';
### ICMP Protocol ##############################################################
$prs[0]{'prs'}[2]{'id'}='icmp';
$prs[0]{'prs'}[2]{'di'}='1';
$prs[0]{'prs'}[2]{'name'}='Internet Control Message Protocol';
$prs[0]{'prs'}[2]{'size'}='8';
$prs[0]{'prs'}[2]{'off'}='9';
$prs[0]{'prs'}[2]{'offs'}='1';
$prs[0]{'prs'}[2]{'pload'}='0';
### RAW Protocol ###############################################################
$prs[0]{'prs'}[3]{'id'}='raw';
$prs[0]{'prs'}[3]{'di'}='255';
$prs[0]{'prs'}[3]{'name'}='RAW Protocol';
$prs[0]{'prs'}[3]{'size'}='32'; # adjust size for raw proto here.
$prs[0]{'prs'}[3]{'off'}='9';
$prs[0]{'prs'}[3]{'offs'}='1';
$prs[0]{'prs'}[3]{'pload'}='0';
################################################################################
### Third Level Protocols Here ? ###############################################
################################################################################

my(%opts,$dev,$offset,$linkoff,$data,@in,$in,$time,$len,$b);
my(@cprot,@cinc,@coff,@csize);
my($i,$j)=(0,0);

use strict;
use Getopt::Std;

getopts('o:c:',\%opts);

if(!$opts{'o'}) {
 print "-"x78 .
       "\nNetParse v$ver by h1kari, for Perl.\n".
       "[http://www.nfsg.org] (Nightfall Security Group)\n".
       "-"x78 ."\n".
       "Usage: nparse.pl <-o<outfile>>[-c<config>]\n".
       "                 -o: specifies an output to read from.\n".
       "                 -c: specifies a config file to use.\n\n"; exit }
($opts{'o'}) || die "You must specify an outfile.\n";
open(IN,"<$opts{'o'}") || die "Unable to open '$opts{'o'}'..\n";
print "Outfile: Found, Using '$opts{'o'}'..\n";
@in=<IN>; $in=join / /, @in; close IN;
$offset=0;
($dev,$linkoff)=unpack('A4L',substr($in,$offset,8)); $offset+=8;
print "Device: $dev..\n";
print "Link Offset: $linkoff..\n";

if($opts{'c'}) {
 open(IN,"<$opts{'c'}");
 while(<IN>) {
  if($_=~/^prot=([^\s]+)\s?([^\n]+)?$/) { $cprot[$i]=$1;
   print "Looking for $cprot[$i] packets";
   if($2=~/^inc='([^']+)'\s?([^\n]+)?$/) { $cinc[$i]=$1;
    $coff[$i]=0; $csize[$i]=$defsize;
    print " containing $cinc[$i]";
    if($2=~/^off=([0-9]+)\s?([^\n]+)?$/) { $coff[$i]=$1;
    print " at offset $coff[$i]";
     if($2=~/^size=([0-9]+)$/) { $csize[$i]=$1;
     print " for $csize[$i] bytes" } } } $i++; print "..\n" } } }

while($offset<length($in)) { my(@tms,$timestamp); my($poff)=$linkoff;
 $b=''; $j=0;
 ($time,$len)=unpack('LL',substr($in,$offset,8)); $offset+=8;
 @tms=localtime $time; $timestamp=" -= [$tms[2]:$tms[1]:$tms[0] ".
  ($tms[4]+1) ."/$tms[3]/$tms[5]] =- ";
 $data=substr($in,$offset,$len); $offset+=$len;
 $b.="\n".
  "|"x((80-length($timestamp))/2) .$timestamp. "|"x((81-length($timestamp))/2);
 printinfo(substr($data,0,$linkoff),"Hardware Link Protocol");
 parsedata($data,"\@prs",$poff);
 $b.="\nLength: $len\n\n";
 if($j) { print $b } }

sub parsedata { my($packet,$prot,$poff)=@_; my($proto,@pros,$prntd,$num);
 $prntd=0;
 eval "\@pros=$prot";
 if(scalar(@pros)==1) { eval "\@pros=\@\{ $prot \}" }
 for(0..scalar(@pros)-1) { my($temp);
  ($pros[$_]{'offs'}<2) && ($temp='c') || ($temp='n');
  $proto=unpack($temp,substr($packet,$linkoff+($pros[$_]{'off'}),
   $pros[$_]{'offs'}));
  if($proto==$pros[$_]{'di'}) { $prntd=1; $num=$_; $prot=~s/\@/\$/g;
   checkprot("$prot\[$num\]",$packet);
   printinfo(substr($packet,$poff,$pros[$num]{'size'}),$pros[$num]{'name'});
   $poff+=$pros[$num]{'size'};
   if(defined($pros[$num]{'prs'})) {
    parsedata($packet,"$prot\[$num\]\{'prs'\}",$poff); return }
   elsif($len-$poff>0) { $prntd=0 } } }
 if(!$prntd) {
  if($pros[$num]{'pload'}!=0) { printdata(substr($packet,$poff,$len-$poff)) }
  else { printinfo(substr($packet,$poff,$len-$poff),"Unknown Protocol") } }
 return }

sub checkprot { my($prot); my($proto,$packet)=@_; ($i==0 && ($j=1)) && return;
 eval("\$prot=$proto\{'id'\}");
 for(0..$i-1) {
  if($prot eq $cprot[$_]) { if(!$cinc[$_]) { $j=1; return }
   if(substr($packet,$linkoff+$coff[$_],$csize[$_])=~/$cinc[$_]/i) {
    $j=1; return } } } return }

sub printinfo { my($packet,$name)=@_; my($count)=0;
 $b.="\n$name:\n";
 for(0..length($packet)-1) { my($token,$bit,$hex,$ord,$string); $count++;
  $token=substr($packet,$_,1); 
  $bit=unpack('B8',$token);
  $hex=unpack('H2',$token);
  $ord=ord($token);
  $string="$bit 0x$hex $ord";
  $b.="$string". " "x(20-length($string));
  if($count==4 || $_==length($packet)-1) { $count=0; $b.="\n" } } return }

sub printdata { my($packet)=$_[0]; my($count)=0;
 $b.="\nPayload:\n";
 for(0..length($packet)-1) { my($token,$hex,$ord);
  $token=substr($packet,$_,1);
  $hex=unpack('H2',$token); $hex="\\x".$hex;
  $ord=ord($token);
  if($ord>=32 && $ord<=126 || ($ord==10)) {
   if($count+1>80) { $b.="\n"; $count=0 }
   ($ord==10) && ($count=0); $b.=$token; $count++ }
  else { 
   if($count+length($hex)>80) { $b.="\n"; $count=0 }
   $b.=$hex; $count+=length($hex) } } $b.="\n"; return }
